#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

#
# AMPL Problem Writer Plugin
#

__all__ = ['Expr_if','ProblemWriter_nl', 'ampl_representation', 'generate_ampl_repn']

from coopr.opt import ProblemFormat, WriterFactory
from coopr.opt.base import *
from coopr.pyomo.base.objective import minimize, maximize
from coopr.pyomo.base import *
from coopr.pyomo.base import expr
from coopr.pyomo.base import external
from coopr.pyomo.base import SymbolMap
from coopr.pyomo.base import intrinsic_functions
from coopr.pyomo.base.var import _VarData, Var, _VarElement
from coopr.pyomo.base.numvalue import NumericConstant, native_numeric_types
from coopr.pyomo.base.param import _ParamData
from coopr.pyomo.base import var
from coopr.pyomo.base import param
from coopr.pyomo.base import numvalue
from coopr.pyomo.base.suffix import active_export_suffix_generator
from pyutilib.component.core import alias
from pyutilib.misc import PauseGC

import itertools
import logging
import operator
import os
import time
import sys

from six import itervalues, iteritems, iterkeys, StringIO
from six.moves import xrange, zip

logger = logging.getLogger('coopr.pyomo')

intrinsic_function_operators = {\
'log':    'o43',\
'log10':  'o42',\
'sin':    'o41',\
'cos':    'o46',
'tan':    'o38',\
'sinh':   'o40',\
'cosh':   'o45',\
'tanh':   'o37',\
'asin':   'o51',\
'acos':   'o53',\
'atan':   'o49',\
'exp':    'o44',\
'sqrt':   'o39',\
'asinh':  'o50',\
'acosh':  'o52',\
'atanh':  'o47',\
'pow':    'o5',\
'abs':    'o15'}

class Expr_if(expr.Expression):
    """An object that defines a dynamic if-then-else expression"""

    __slots__ = ('_if','_then','_else')

    def __init__(self, IF=None, THEN=None, ELSE=None):
        """Constructor"""
        Expression.__init__(self, None)
        
        self._if = numvalue.as_numeric(IF)
        self._then = numvalue.as_numeric(THEN)
        self._else = numvalue.as_numeric(ELSE)

    def __getstate__(self):
        result = Expression.__getstate__(self)
        for i in Expr_if.__slots__:
            result[i] = getattr(self, i)
        return result

    def cname(self):
        return "Expr_if"

    def is_constant(self):
        if self._if.is_constant():
            if bool(self._if) is True:
                return self._then.is_constant()
            else:
                return self._else.is_constant()
        else:
            # Even if the branching point is not constant, if both
            # branches are constant and equal then this expression
            # is constant
            if self._then.is_constant() and self._else.is_constant():
                return value(self._then) == value(self._else)

    def is_fixed(self):
        if self._if.is_fixed():
            if bool(self._if) is True:
                return self._then.is_fixed() 
            else:
                return self._else.is_fixed()
        else:
            # Even if the branching point is not fixed, if both
            # branches are fixed and equal then this expression
            # is fixed
            if self._then.is_fixed() and self._else.is_fixed():
                return value(self._then) == value(self._else)

    def polynomial_degree(self):
        if self._if.is_fixed():
            if bool(self._if) is True:
                return self._then.polynomial_degree()
            else:
                return self._else.polynomial_degree()

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            ostream.write("Expr_if( if=( ")
            if self._if.is_expression():
                self._if.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
            else:
                ostream.write(str(self._if)+' ')
            ostream.write(") ")
            ostream.write(", then=( ")
            if self._then.is_expression():
                self._then.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
            else:
                ostream.write(str(self._then)+' ')
            ostream.write(") ")
            ostream.write(", else=( ")
            if self._else.is_expression():
                self._else.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
            else:
                ostream.write(str(self._else)+' ')
            ostream.write(") ")
            if eol_flag==True:
                ostream.write("\n")

    def __copy__(self):
        """Clone this object using the specified arguments"""
        tmp = self.__class__()
        tmp._if = clone_expression(self._if)
        tmp._then = clone_expression(self._then)
        tmp._else = clone_expression(self._else)
        return tmp

    def __call__(self, exception=True):
        """Evaluate the expression"""
        if bool(self._if) is True:
            return value(self._then)
        else:
            return value(self._else)

class _trivial_labeler(object):

    def __init__(self,start_id=0):
        self._id = start_id

    def __call__(self, obj):
        self._id += 1
        return self._id

class ModelSOS(object):

    class AmplSuffix(object):

        def __init__(self,name):
            self.name = name
            self.ids = []
            self.vals = []

        def add(self,idx,val):
            if idx in self.ids:
                raise RuntimeError("The NL file format does not support multiple nonzero "\
                                    "values for a single component and suffix. \n"\
                                    "Suffix Name:  %s\n"\
                                    "Component ID: %s\n" % (self.name, idx))
            else:
                self.ids.append(idx)
                self.vals.append(val)
                    
        def genfilelines(self):
            base_line = "{0} {1}\n"
            return [base_line.format(idx, val) for idx, val in zip(self.ids,self.vals) if val != 0]

        def is_empty(self):
            return not bool(len(self.ids))

    def __init__(self,ampl_var_id,symbol_map,labeler):
        self.ampl_var_id = ampl_var_id
        self.symbol_map = symbol_map
        self.labeler = labeler
        self.sosno = self.AmplSuffix('sosno')
        self.ref = self.AmplSuffix('ref')
        self.block_cntr = 0

    def count_constraint(self,con,level,index=None):
        
        ampl_var_id = self.ampl_var_id
        symbol_map = self.symbol_map
        labeler = self.labeler
        sos_vars = con.sos_vars()

        # Identifies the the current set from others in the NL file
        self.block_cntr += 1

        # If SOS1, the identifier must be positive
        # if SOS2, the identifier must be negative
        sign_tag = 1
        if level == 2:
            sign_tag = -1

        if index is None:
            tmpSet = con.sos_set()
        else:
            tmpSet = con.sos_set()[index]

        # Get all the variables
        cntr = 0
        for x in tmpSet:
            ID = ampl_var_id[symbol_map.getSymbol( sos_vars[x], labeler )]
            self.sosno.add(ID,self.block_cntr*sign_tag)
            # We need to weight each variable
            # For now we just increment a counter
            if cntr > 0:
                # We only need to report variables with weight greater than zero
                # Zero weight is assumed for variables with no weight reported.
                self.ref.add(ID,cntr)
            cntr += 1


class ProblemWriter_nl(AbstractProblemWriter):

    alias(str(ProblemFormat.nl))

    def __init__(self):
        self.ampl_var_id = {}
        self.ampl_con_id = {}
        self.ampl_obj_id = {}
        AbstractProblemWriter.__init__(self, ProblemFormat.nl)

    def __call__(self, model, filename, solver_capability, symbolic_solver_labels, **kwds):
        # Pause the GC for the duration of this method
        suspend_gc = PauseGC()
        if filename is None:
            filename = model.cname() + ".nl"
        OUTPUT=open(filename,"w")
        symbol_map = self._print_model_NL(model, OUTPUT, solver_capability, symbolic_solver_labels, verbose=False, gen_obj_ampl_repn=True, gen_con_ampl_repn=True)
        OUTPUT.close()
        return filename, symbol_map

    def _get_bound(self, exp):
        if exp.is_fixed():
            return exp()
        else:
            raise ValueError("ERROR: nonconstant bound: " + str(exp))
            return None

        exp_type = type(exp)
        # Params & Constants are fixed, so are handled above
        #if exp_type is NumericConstant:
        #    return exp.value
        #elif exp_type is _ParamData:
        #    return exp.value
        if exp_type is expr._SumExpression:
            sum_expr_value=0.0
            for i in xrange(len(exp._args)):
                sum_expr_value += self._get_bound(exp._args[i])*exp._coef[i]
            return sum_expr_value
        elif exp_type is expr._ProductExpression:
            prod_expr_value=1.0
            for i in xrange(len(exp._numerator)):
                prod_expr_value *= float(self._get_bound(exp._numerator[i]))
            for i in xrange(len(exp._denominator)):
                prod_expr_value /= float(self._get_bound(exp._denominator[i]))
            return prod_expr_value
        else:
            raise ValueError("ERROR: nonconstant bound: " + str(exp))
            return None

    def _print_nonlinear_terms_NL(self, OUTPUT, exp, symbol_map, labeler):

        exp_type = type(exp)
        # JDS: check list first so that after this, we know that exp
        # must be some form of NumericValue
        if exp_type is list:
            # this is an implied summation of expressions (we did not create a new sum expression for efficiency)
            # this should be a list of tuples where [0] is the coeff and [1] is the expr to write
            n = len(exp)
            if n > 2:
                # sum
                OUTPUT.write( "o54\n%d\n"%(n) )
                for i in xrange(0,n):
                    assert(exp[i].__class__ is tuple)
                    coef = exp[i][0]
                    child_exp = exp[i][1]
                    if coef != 1.0:
                        # *
                        OUTPUT.write( "o2\nn{0!r}\n".format(coef) )
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
            else:
                for i in xrange(0,n):
                    assert(exp[i].__class__ is tuple)
                    coef = exp[i][0]
                    child_exp = exp[i][1]
                    if i != n-1:
                        # need the + op if it is not the last entry in the list
                        # +
                        OUTPUT.write( "o0\n" )
                    if coef != 1.0:
                        # *
                        OUTPUT.write( "o2\nn{0!r}\n".format(coef) )
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
        elif exp.is_expression():
            if exp_type is expr._SumExpression:
                n = len(exp._args)
                if n > 2:
                    # sum
                    OUTPUT.write( "o54\n%d\n"%(n) )
                    if exp._const != 0.0:
                        # +
                        OUTPUT.write( "o0\nn{0!r}\n".format(exp._const) )
                    for i in xrange(0,n):
                        if exp._coef[i] != 1:
                            # *
                            OUTPUT.write( "o2\nn{0!r}\n".format(exp._coef[i]) )
                        self._print_nonlinear_terms_NL(OUTPUT, exp._args[i], symbol_map, labeler)
                else:
                    if exp._const != 0.0:
                        # +
                        OUTPUT.write( "o0\nn{0!r}\n".format(exp._const) )
                    for i in xrange(0,n-1):
                        # +
                        OUTPUT.write( "o0\n" )
                        if exp._coef[i] != 1.0:
                            # *
                            OUTPUT.write( "o2\nn{0!r}\n".format(exp._coef[i]) )
                        self._print_nonlinear_terms_NL(OUTPUT, exp._args[i], symbol_map, labeler)
                    if exp._coef[n-1] != 1.0:
                        # *
                        OUTPUT.write( "o2\nn{0!r}\n".format(exp._coef[n-1]) )
                    self._print_nonlinear_terms_NL(OUTPUT, exp._args[n-1], symbol_map, labeler)
    
            elif exp_type is expr._ProductExpression:
                denom_exists = False
                if len(exp._denominator) == 0:
                    pass
                else:
                    # /
                    OUTPUT.write( "o3\n" )
                    denom_exists = True
                if exp._coef != 1.0:
                    # *
                    OUTPUT.write( "o2\nn{0!r}\n".format(exp._coef) )
                if len(exp._numerator) == 0:
                    OUTPUT.write( "n1\n" )
                # print out the numerator
                child_counter = 0
                max_count = len(exp._numerator)-1
                for child_exp in exp._numerator:
                    if child_counter < max_count:
                        # *
                        OUTPUT.write( "o2\n" )
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
                    child_counter += 1
                if denom_exists:
                    # print out the denominator
                    child_counter = 0
                    max_count = len(exp._denominator)-1
                    for child_exp in exp._denominator:
                        if child_counter < max_count:
                            # *
                            OUTPUT.write( "o2\n" )
                        self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
                        child_counter += 1
    
            elif exp_type is external._ExternalFunctionExpression:
                OUTPUT.write("f%d %d\n" % (
                        self.external_byFcn[exp._fcn._function][1], 
                        len(exp._args) ) )
                for arg in exp._args:
                    if isinstance(arg, basestring):
                        OUTPUT.write("h%d:%s\n" % (len(arg), arg))
                    else:
                        self._print_nonlinear_terms_NL(OUTPUT, arg, symbol_map, labeler)
            elif isinstance(exp, expr._IntrinsicFunctionExpression):
                intr_expr = intrinsic_function_operators.get(exp.cname(),None)
                if intr_expr is not None:
                    OUTPUT.write( intr_expr+"\n" )
                else:
                    logger.error("Unsupported intrinsic function ({0})", exp.cname(True))
                    raise TypeError("ASL writer does not support '{0}' expressions"
                                   .format(exp.cname(True)))
                    
                for child_exp in exp._args:
                    self._print_nonlinear_terms_NL(OUTPUT, child_exp, symbol_map, labeler)
            elif exp_type is Expr_if:
                OUTPUT.write("o35\n")
                self._print_nonlinear_terms_NL(OUTPUT, exp._if, symbol_map, labeler)
                self._print_nonlinear_terms_NL(OUTPUT, exp._then, symbol_map, labeler)
                self._print_nonlinear_terms_NL(OUTPUT, exp._else, symbol_map, labeler)
            elif exp_type is expr._InequalityExpression:
                len_args = len(exp._args)
                assert len_args in [2,3]
                left = exp._args[0]
                middle = exp._args[1]
                right = None
                if len_args == 3:
                    right = exp._args[2]
                    OUTPUT.write("o21\n") # and
                if exp._strict[0]:
                    OUTPUT.write("o22\n")
                else:
                    OUTPUT.write("o23\n")
                self._print_nonlinear_terms_NL(OUTPUT, left, symbol_map, labeler)
                self._print_nonlinear_terms_NL(OUTPUT, middle, symbol_map, labeler)
                if not right is None:
                    if exp._strict[1]:
                        OUTPUT.write("o22\n")
                    else:
                        OUTPUT.write("o23\n")
                    self._print_nonlinear_terms_NL(OUTPUT, middle, symbol_map, labeler)
                    self._print_nonlinear_terms_NL(OUTPUT, right, symbol_map, labeler)
            elif exp_type is expr._EqualityExpression:
                OUTPUT.write("o24\n")
                self._print_nonlinear_terms_NL(OUTPUT, exp._args[0], symbol_map, labeler)
                self._print_nonlinear_terms_NL(OUTPUT, exp._args[1], symbol_map, labeler)
            else:
                raise ValueError("Unsupported expression type (%s) in _print_nonlinear_terms_NL" % exp_type)

        elif isinstance(exp,var._VarData) and not exp.is_fixed():
            variable_label = symbol_map.getSymbol(exp, labeler)
            OUTPUT.write( "v{0}\n".format(self.ampl_var_id[variable_label]) )
        elif isinstance(exp,param._ParamData):
            OUTPUT.write( "n{0!r}\n".format(value(exp)) )
        elif isinstance(exp,NumericConstant) or exp.is_fixed():
            OUTPUT.write( "n{0!r}\n".format(value(exp)) )
        else:
            raise ValueError("Unsupported expression type (%s) in _print_nonlinear_terms_NL" % exp_type)
        
    
    def _print_model_NL(self, model, OUTPUT, solver_capability, symbolic_solver_labels, verbose=False, gen_obj_ampl_repn=True, gen_con_ampl_repn=True):

        # maps NL variables to the "real" variable names in the problem.
        # it's really NL variable ordering, as there are no variable names
        # in the NL format. however, we by convention make them go from
        # x0 upward.

#        start_overall_time = time.clock()

        # because NL preprocessing occurs during the write process, we don't have direct access to
        # this method through the solve() method of a solver plugin. and keyword propagation is too
        # complex to think that would be a good solution. so, we can optionally tag the input
        # instance with the preprocessing keywords.
        if hasattr(model, "gen_obj_ampl_repn"):
            gen_obj_ampl_repn = getattr(model, "gen_obj_ampl_repn")
            delattr(model, "gen_obj_ampl_repn")
        if hasattr(model, "gen_con_ampl_repn"):
            gen_con_ampl_repn = getattr(model, "gen_con_ampl_repn")
            delattr(model, "gen_con_ampl_repn")
        if hasattr(model, "_NL_symbol_map"):
            # The main purpose of this code is so that
            # PYSP can rely on "gen_con_ampl_repn=False" (from above)
            # If we create a new symbol map for the objective (which may
            # have new terms) the symbol applied to the terms in the objective
            # may not match those used for the old constraint ampl_repns.
            # This is now the case since the labeler now longer uses fully
            # qualified names for labels, but instead uses an incremented (in 
            # the order encountered) integer id.
            symbol_map = model._NL_symbol_map
            symbol_map.clearAliases()
            assert model._NL_symbol_map.instance() is model
            # start the labeling counter at what it stopped at for the old symbol map
            labeler = _trivial_labeler(start_id=max(iterkeys(symbol_map.bySymbol)))
        else:
            symbol_map = SymbolMap(model)
            labeler = _trivial_labeler()
            model._NL_symbol_map = symbol_map
        # Speeds up calling cname on every component when
        # writing .row and .col files (when symbolic_solver_labels is True)
        name_labeler = NameLabeler()

        #
        # Collect statistics
        #
        self_ampl_var_id = self.ampl_var_id = {}
        self_ampl_con_id = self.ampl_con_id = {}
        self_ampl_obj_id = self.ampl_obj_id = {}

        Vars_dict = dict() # will be the entire list of all vars used in the problem
        Objectives_dict = dict() # will be the entire list of all objective used in the problem (size=1)
        Constraints_dict = dict() # will be the entire list of all constraints used in the problem

        # linear variables
        LinearVars = set()
        LinearVarsInt = set()
        LinearVarsBool = set()

        # Tabulate the External Function definitions
        self.external_byFcn = {}
        external_Libs = set()
        for block in model.all_blocks():
            for fcn in active_subcomponents_generator(block,ExternalFunction):
                if fcn._function in self.external_byFcn and \
                        self.external_byFcn[fcn._function][0]._library != fcn._library:
                    raise RuntimeError(
                        "The same external function name (%s) is associated "
                        "with two different libraries (%s through %s, and %s "
                        "through %s).  The ASL solver will fail to link "
                        "correctly." %
                        ( fcn._function, 
                          self.external_byFcn[fcn._function]._library,
                          self.external_byFcn[fcn._function]._library.cname(True),
                          fcn._library,
                          fcn.cname(True) ) )
                self.external_byFcn[fcn._function] = (fcn, len(self.external_byFcn))
                external_Libs.add(fcn._library)
        if external_Libs:
            os.environ["COOPR_AMPLFUNC"] = "\n".join(sorted(external_Libs))
        elif "COOPR_AMPLFUNC" in os.environ:
            del os.environ["COOPR_AMPLFUNC"]


#        print "Generate objective representation:",
#        start_time = time.clock()

        #
        # Count number of objectives and build the ampl_repns
        #
        n_objs = 0
        n_nonlinear_objs = 0
        ObjVars = set()
        ObjNonlinearVars = set()
        ObjNonlinearVarsInt = set()
        active_objective = None
        # We only check for an objective on the top level model
        # Create a suffix component for the ampl_repn
        obj_ampl_repn_suffix = model.subcomponent("ampl_repn")
        if (obj_ampl_repn_suffix is None) or (not obj_ampl_repn_suffix.type() is Suffix) or (not obj_ampl_repn_suffix.active is True):
            obj_ampl_repn_suffix = model.ampl_repn = Suffix(direction=Suffix.LOCAL,datatype=None)

        for active_objective in active_subcomponents_data_generator(model,Objective):
            if gen_obj_ampl_repn is True:
                ampl_repn = generate_ampl_repn(active_objective.expr, symbol_map, labeler)
                obj_ampl_repn_suffix.setValue(active_objective, ampl_repn)
            else:
                ampl_repn = obj_ampl_repn_suffix.getValue(active_objective)

            Vars_dict.update(ampl_repn._linear_terms_var)
            Vars_dict.update(ampl_repn._nonlinear_vars)
            
            LinearVars.update(ampl_repn._linear_terms_var)
            ObjNonlinearVars.update(ampl_repn._nonlinear_vars)
            
            ObjVars.update(ampl_repn._linear_terms_var)
            ObjVars.update(ampl_repn._nonlinear_vars)
            n_objs += 1
            if ampl_repn.is_nonlinear():
                n_nonlinear_objs += 1
        
        # I don't think this is necessarily true for the entire code base, 
        # but seeing has how its never been tested we should go ahead and
        # raise an exception
        if n_objs > 1:
            raise ValueError("The NL writer has detected multiple active objective functions, but currently only handles a single objective.")
        else:
            # ***IMPORTANT:
            # only at this point is it safe to alias the single 
            # active objective as __default_objective__. This is 
            # important for loading objective suffixes into the
            # results object in the .sol reader and also for the 
            # ASL hack inside pyomo/scripting/util.process_results
            # where the objective value needs to be manually 
            # calculated and then placed in the results object.
            label = symbol_map.getSymbol(active_objective, labeler)
            symbol_map.alias(active_objective, "__default_objective__")
            Objectives_dict[label] = (active_objective, obj_ampl_repn_suffix.getValue(active_objective))
            self_ampl_obj_id[label] = 0

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Generate constraint repns:",
#        start_time = time.clock()

        #
        # Count number of constraints and build the ampl_repns
        #
        n_ranges = 0
        n_single_sided_ineq = 0
        n_equals = 0
        n_nonlinear_constraints = 0
        ConNonlinearVars = set()
        ConNonlinearVarsInt = set()
        nnz_grad_constraints = 0
        constraint_bounds_dict = {}
        nonlin_con_order_list = []
        lin_con_order_list = []        

        for block in model.all_blocks():
            all_repns = list()
            # Create a suffix component for the ampl_repn
            block_ampl_repn_suffix = block.subcomponent("ampl_repn")
            if (block_ampl_repn_suffix is None) or (not block_ampl_repn_suffix.type() is Suffix) or (not block_ampl_repn_suffix.active is True):
                block_ampl_repn_suffix = block.ampl_repn = Suffix(direction=Suffix.LOCAL,datatype=None)
            for constraint_data in active_subcomponents_data_generator(block,Constraint):
                
                con_name = symbol_map.getSymbol( constraint_data, labeler )
                if gen_con_ampl_repn is True:
                    ampl_repn = generate_ampl_repn(constraint_data.body, symbol_map, labeler)
                    all_repns.append((constraint_data,ampl_repn))
                else:
                    ampl_repn = block_ampl_repn_suffix.getValue(constraint_data)

                linear_vars = ampl_repn._linear_terms_var
                nonlinear_vars = ampl_repn._nonlinear_vars
                
                ### GAH: Even if this is fixed, it is still useful to write out these types of constraints
                ###      (trivial) as a feasibility check for fixed variables, in which case the solver
                ###      will pick up on the model infeasibility.
                ### **NOTE: If we change this loop to start skipping fixed constraints, then the "J" section must
                ###         be updated to do so as well since the constraint id is calculated on the fly using
                ###         a cntr.
                #if ampl_repn.is_fixed():
                #    continue
                    
                if ampl_repn.is_nonlinear():
                    nonlin_con_order_list.append(con_name)
                    n_nonlinear_constraints += 1
                else:
                    lin_con_order_list.append(con_name)

                Constraints_dict[con_name] = (constraint_data, ampl_repn)
                
                # Merge variables from this expression
                Vars_dict.update(linear_vars)
                Vars_dict.update(nonlinear_vars)
                # These are sets
                LinearVars.update(linear_vars)
                ConNonlinearVars.update(nonlinear_vars)
                    
                nnz_grad_constraints += len(set(linear_vars).union(nonlinear_vars))
    
                L = None
                U = None
                if constraint_data.lower is not None: L = self._get_bound(constraint_data.lower)
                if constraint_data.upper is not None: U = self._get_bound(constraint_data.upper)
                offset = ampl_repn._constant
                #if constraint_data._equality is True:
                #    assert L == U
                if L == U:
                    constraint_bounds_dict[con_name] = "4 {0!r}\n".format(L-offset)
                    n_equals += 1
                elif L is None:
                    constraint_bounds_dict[con_name] = "1 {0!r}\n".format(U-offset)
                    n_single_sided_ineq += 1
                elif U is None:
                    constraint_bounds_dict[con_name] = "2 {0!r}\n".format(L-offset)
                    n_single_sided_ineq += 1
                elif (L > U):
                    msg = 'Constraint {0}: lower bound greater than upper' \
                        ' bound ({1} > {2})'
                    raise ValueError(msg.format(con_name, str(L), str(U)))
                else:
                    constraint_bounds_dict[con_name] = "0 {0!r} {1!r}\n".format(L-offset, U-offset)
                    # double sided inequality
                    # both are not none and they are valid
                    n_ranges += 1
            block_ampl_repn_suffix.updateValues(all_repns)

        # create the ampl constraint ids
        self_ampl_con_id.update((name,idx) for idx,name in 
                           enumerate(itertools.chain(nonlin_con_order_list,lin_con_order_list)))

#        end_time = time.clock()
#        print (end_time - start_time)


#        print "Partitioning Variable Types:",
#        start_time = time.clock()

        LinearVars = LinearVars.difference(ObjNonlinearVars)
        LinearVars = LinearVars.difference(ConNonlinearVars)

        ### There used to be an if statement here for the following code block 
        ### checking model.statistics.num_binary_vars was greater than zero.
        ### To this day, I don't know how it worked.
        ### TODO: Figure out why
        ###############
        for var_name in LinearVars:
            var = Vars_dict[var_name]
            var_type = type(var.domain)
            if var_type is IntegerSet:
                LinearVarsInt.add(var_name)
            elif var_type is BooleanSet:
                L = var.lb
                U = var.ub
                if L is None or U is None:
                    raise ValueError("Variable " + str(var.cname(True)) +\
                                     "is binary, but does not have lb and ub set")
                elif value(L) == 0 and value(U) == 1:
                    LinearVarsBool.add(var_name)
                else:
                    LinearVarsInt.add(var_name)
        LinearVars.difference_update(LinearVarsInt)
        LinearVars.difference_update(LinearVarsBool)

        for var_name in ObjNonlinearVars:
            var_type = type(Vars_dict[var_name].domain)
            if var_type is IntegerSet or var_type is BooleanSet:
                ObjNonlinearVarsInt.add(var_name)
        ObjNonlinearVars.difference_update(ObjNonlinearVarsInt)
        for var_name in ConNonlinearVars:
            var_type = type(Vars_dict[var_name].domain)
            if var_type is IntegerSet or var_type is BooleanSet:
                ConNonlinearVarsInt.add(var_name)
        ConNonlinearVars.difference_update(ConNonlinearVarsInt)
        ##################

        Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVars.intersection(ConNonlinearVars)
        Discrete_Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVarsInt.intersection(ConNonlinearVarsInt)
        ObjNonlinearVars = ObjNonlinearVars.difference(Nonlinear_Vars_in_Objs_and_Constraints)
        ConNonlinearVars = ConNonlinearVars.difference(Nonlinear_Vars_in_Objs_and_Constraints)
        ObjNonlinearVarsInt = ObjNonlinearVarsInt.difference(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        ConNonlinearVarsInt = ConNonlinearVarsInt.difference(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        
        # put the ampl variable id into the variable
        full_var_list = []
        full_var_list.extend(sorted(Nonlinear_Vars_in_Objs_and_Constraints))
        full_var_list.extend(sorted(Discrete_Nonlinear_Vars_in_Objs_and_Constraints))
        idx_nl_both = len(full_var_list)
        #
        full_var_list.extend(sorted(ConNonlinearVars))
        full_var_list.extend(sorted(ConNonlinearVarsInt))
        idx_nl_con = len(full_var_list)
        #
        full_var_list.extend(sorted(ObjNonlinearVars))
        full_var_list.extend(sorted(ObjNonlinearVarsInt))
        idx_nl_obj = len(full_var_list)
        #
        full_var_list.extend(sorted(LinearVars))
        full_var_list.extend(sorted(LinearVarsBool))
        full_var_list.extend(sorted(LinearVarsInt))

        if (idx_nl_obj == idx_nl_con):
            idx_nl_obj = idx_nl_both

        # create the ampl variable ids
        self_ampl_var_id.update( (name,idx) for (idx,name) in enumerate(full_var_list))

        ##################################################
        ############ # LQM additions start here ##########
        ##################################################

        ###                                                                                                    
        # Generate file containing matrices for continuity variables between blocks.                            
        # This checks for a LOCAL suffix 'lqm' on variables.
        ###
        lqm_suffix = model.subcomponent("lqm")
        if (lqm_suffix is None) or (not (lqm_suffix.type() is Suffix)) or \
           (not (lqm_suffix.active is True)) or (not (lqm_suffix.getDirection() is Suffix.LOCAL)):
            lqm_suffix = None
        if lqm_suffix is not None:
            lqm_var_column_ids = []
            for var in itervalues(Vars_dict):
                lqm = lqm_suffix.getValue(var)
                if lqm != -1:
                    # store the column index (and translate to one-based indexing
                    lqm_var_column_ids.append(self_ampl_var_id[symbol_map.getSymbol(var,labeler)]+1)

            num_lqm_vars = len(lqm_var_column_ids)
            if num_lqm_vars > 0:
                lqm_output_name = OUTPUT.name.split('.')[0]+".lqm"
                LQM_OUTPUT = open(lqm_output_name,'w')
                LQM_OUTPUT.write( "Matrix Li\n" )
                LQM_OUTPUT.write( "%d\n"%(len(full_var_list)) )
                LQM_OUTPUT.write( "%d\n"%(num_lqm_vars) )
                LQM_OUTPUT.write( "%d\n"%(num_lqm_vars) )
                # The matrix uses one based indexing
                for row_id, col_id in enumerate(lqm_var_column_ids,1):
                    LQM_OUTPUT.write( "%d %d 1\n" % (row_id, col_id) )
                LQM_OUTPUT.write( "Matrix Qi\n" )
                LQM_OUTPUT.write("%d\n" % (num_lqm_vars) )
                LQM_OUTPUT.write("%d\n" % (num_lqm_vars) )
                LQM_OUTPUT.write("%d\n" % (num_lqm_vars) )
                # one based indexing
                for counter in xrange(1,num_lqm_vars+1):
                    LQM_OUTPUT.write("%d %d -1\n" % (counter, counter) )
                LQM_OUTPUT.close()

        ##################################################
        ############ # LQM additions end here ############
        ##################################################

#        end_time = time.clock()
#        print (end_time - start_time)

        colfilename = None
        if OUTPUT.name.endswith('.nl'):
            colfilename = OUTPUT.name.replace('.nl','.col')
        else:
            colfilename = OUTPUT.name+'.col'
        if symbolic_solver_labels is True:
            colf = open(colfilename,'w')
            for var_name in full_var_list:
                colf.write(name_labeler(Vars_dict[var_name])+'\n')
            colf.close()
        
#        print "Printing NL file header:",
#        start_time = time.clock()
        #
        # Print Header
        #
        # LINE 1
        #
        OUTPUT.write("g3 1 1 0\t# problem {0}\n".format(model.cname()) )
        #
        # LINE 2
        #
        OUTPUT.write( " {0} {1} {2} {3} {4} \t# vars, constraints, objectives, general inequalities, equalities\n" \
                          .format(len(full_var_list), n_single_sided_ineq+n_ranges+n_equals, n_objs, n_ranges, n_equals) )
        #
        # LINE 3
        #
        OUTPUT.write( " {0} {1} \t# nonlinear constraints, objectives\n".format(n_nonlinear_constraints, n_nonlinear_objs) )
        #
        # LINE 4
        #
        OUTPUT.write( " 0 0\t# network constraints: nonlinear, linear\n" )
        #
        # LINE 5
        #
        OUTPUT.write( " {0} {1} {2} \t# nonlinear vars in constraints, objectives, both\n" \
                          .format(idx_nl_con, idx_nl_obj, idx_nl_both) )

        #
        # LINE 6
        #
        OUTPUT.write( " 0 {0} 0 1\t# linear network variables; functions; arith, flags\n".format(len(self.external_byFcn)) )
        #
        # LINE 7
        #
        n_int_nonlinear_b = len(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        n_int_nonlinear_c = len(ConNonlinearVarsInt)
        n_int_nonlinear_o = len(ObjNonlinearVarsInt) 
        OUTPUT.write( " {0} {1} {2} {3} {4} \t# discrete variables: binary, integer, nonlinear (b,c,o)\n" \
           .format(len(LinearVarsBool), len(LinearVarsInt), n_int_nonlinear_b, n_int_nonlinear_c, n_int_nonlinear_o) )
        #
        # LINE 8
        #
        # objective info computed above
        OUTPUT.write( " {0} {1} \t# nonzeros in Jacobian, obj. gradient\n".format(nnz_grad_constraints, len(ObjVars)) )
        #
        # LINE 9
        #
        OUTPUT.write( " 0 0\t# max name lengths: constraints, variables\n" )
        #
        # LINE 10
        #
        OUTPUT.write( " 0 0 0 0 0\t# common exprs: b,c,o,c1,o1\n" )

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing constraints:",
#        start_time = time.clock()


        #
        # "F" lines
        #
        for fcn, fid in sorted(itervalues(self.external_byFcn), key=operator.itemgetter(1)):
            OUTPUT.write( "F%d 1 -1 %s\n" % ( fid, fcn._function ))

        #
        # "S" lines
        #

        # Tranlate the SOSConstraint component into ampl suffixes
        # SOS constraints - largely taken from cpxlp.py so updates there,
        # should be applied here
        # TODO: Allow users to specify the variables coefficients for custom
        # branching/set orders - refer to cpxlp.py
        sos1 = solver_capability("sos1")
        sos2 = solver_capability("sos2")
        modelSOS = ModelSOS(self_ampl_var_id,symbol_map,labeler)
        if sos1 or sos2:
            for block in model.all_blocks():
                for con in itervalues(block.active_subcomponents(SOSConstraint)):
                    level = con.sos_level()
                    if (level == 1 and sos1) or (level == 2 and sos2):
                        masterIndex = con.sos_set_set()
                        if None in masterIndex:
                            modelSOS.count_constraint(con, level)
                        else:
                            for index in masterIndex:
                                modelSOS.count_constraint(con, level, index)
        var_sosno_suffix = modelSOS.sosno
        var_ref_suffix = modelSOS.ref
        sosconstraint_sosno_vals = set(var_sosno_suffix.vals)

        # Translate the rest of the Pyomo Suffix components
        suffix_header_line = "S{0} {1} {2}\n"
        suffix_line = "{0} {1!r}\n"
        var_tag = 0
        con_tag = 1
        obj_tag = 2
        prob_tag = 3
        suffix_dict = dict(active_export_suffix_generator(model))
        has_dual_initials = False
        if not ('sosno' in suffix_dict):
            # We still need to write out the SOSConstraint suffixes
            # even though these may have not been "declared" on the model
            s_lines = var_sosno_suffix.genfilelines()
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(var_tag,len_s_lines,'sosno'))
                [OUTPUT.write(line) for line in s_lines]
        else:
            # I am choosing not to allow a user to mix the use of the Pyomo
            # SOSConstraint component and manual sosno declarations within
            # a single model. I initially tried to allow this but the
            # var section of the code below blows up for two reason. (1)
            # we have to make sure that the sosno suffix is not defined
            # twice for the same variable (2) We have to make sure that
            # the automatically chosen sosno used by the SOSConstraint
            # translation does not already match one a user has manually
            # implemented (this would modify the members in an sos set).
            # Since this suffix is exclusively used for defining sos sets,
            # there is no reason a user can not just stick to one method.
            if not var_sosno_suffix.is_empty():
                raise RuntimeError("The Pyomo NL file writer does not allow both manually "\
                                    "declared 'sosno' suffixes as well as SOSConstraint "\
                                    "components to exist on a single model. To avoid this "\
                                    "error please use only one of these methods to define "\
                                    "special ordered sets.")
        if not ('ref' in suffix_dict):
            # We still need to write out the SOSConstraint suffixes
            # even though these may have not been "declared" on the model
            s_lines = var_ref_suffix.genfilelines()
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(var_tag,len_s_lines,'ref'))
                [OUTPUT.write(line) for line in s_lines]
        else:
            # see reason (1) in the paragraph above for why we raise this
            # exception (replacing sosno with ref).
            if not var_ref_suffix.is_empty():
                raise RuntimeError("The Pyomo NL file writer does not allow both manually "\
                                    "declared 'ref' suffixes as well as SOSConstraint "\
                                    "components to exist on a single model. To avoid this "\
                                    "error please use only one of these methods to define "\
                                    "special ordered sets.")
        for suffix_name, suffix in iteritems(suffix_dict): 
            if suffix.getDatatype() not in (float,int,bool,str):
                raise ValueError("The Pyomo NL file writer requires that all active export Suffix components "\
                                  "have a strict datatype equal to one of %s, but Suffix component with "\
                                  "name %s has datatype %s" % ((float,int,bool,str), suffix_name, suffix.getDatatype()))
            if suffix_name == "dual":
                # The NL file format has a special section for dual initializations
                has_dual_initials = True
                continue
            float_tag = 0
            if suffix.getDatatype() is float:
                float_tag = 4
            ################## vars
            s_lines = []
            for var_name, var_component in iteritems(Vars_dict):
                suffix_value = suffix.getNumericValue(var_component)
                if suffix_value is not None:
                    s_lines.append(suffix_line.format(self_ampl_var_id[var_name],suffix_value))
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(var_tag | float_tag, len_s_lines,suffix_name))
                [OUTPUT.write(line) for line in s_lines]
                    
            ################## constraints
            s_lines = []
            for con_name, (con_component, ampl_repn) in iteritems(Constraints_dict):
                suffix_value = suffix.getNumericValue(con_component)
                if suffix_value is not None:
                    s_lines.append(suffix_line.format(self_ampl_con_id[con_name],suffix_value))
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(con_tag | float_tag, len_s_lines,suffix_name))
                [OUTPUT.write(line) for line in s_lines]

            ################## objectives
            s_lines = []
            for obj_name, (obj_component, ampl_repn) in iteritems(Objectives_dict):
                suffix_value = suffix.getNumericValue(obj_component)
                if suffix_value is not None:
                    s_lines.append(suffix_line.format(self_ampl_obj_id[obj_name],suffix_value))
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(obj_tag | float_tag, len_s_lines,suffix_name))
                [OUTPUT.write(line) for line in s_lines]

            ################## problems (in this case the one problem)
            suffix_value = suffix.getNumericValue(model)
            if suffix_value is not None:
                OUTPUT.write(suffix_header_line.format(prob_tag | float_tag, 1,suffix_name))
                OUTPUT.write(suffix_line.format(0,suffix_value))
        del modelSOS

        #
        # "C" lines
        #
        rowfilename = None
        if OUTPUT.name.endswith('.nl'):
            rowfilename = OUTPUT.name.replace('.nl','.row')
        else:
            rowfilename = OUTPUT.name+'.row'
        if symbolic_solver_labels is True:
            rowf = open(rowfilename,'w')

        cu = [0 for i in xrange(len(full_var_list))]
        for con_name in nonlin_con_order_list:
            con_data, ampl_repn = Constraints_dict[con_name]
            con_id = self_ampl_con_id[con_name]
            OUTPUT.write( "C%d\n"%(con_id) )
            if symbolic_solver_labels is True:
                rowf.write( name_labeler(con_data)+"\n" ) 
            symbol_map.alias(con_data, ".c%d"%(con_id))
            self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr, symbol_map, labeler)

            for d in set(ampl_repn._linear_terms_var).union(ampl_repn._nonlinear_vars):
                cu[self_ampl_var_id[d]] += 1

        for con_name in lin_con_order_list:
            con_data, ampl_repn = Constraints_dict[con_name]
            con_id = self_ampl_con_id[con_name]
            con_vars = set(ampl_repn._linear_terms_var)
            for d in con_vars:
                cu[self_ampl_var_id[d]] += 1
            OUTPUT.write( "C%d\n"%(con_id) )
            if symbolic_solver_labels is True:
                rowf.write( name_labeler(con_data)+"\n" )
            symbol_map.alias(con_data, ".c"+str(con_id))
            OUTPUT.write( "n0\n" )

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing objective:",
#        start_time = time.clock()
        #
        # "O" lines
        #
        for obj_name, (obj, ampl_repn) in iteritems(Objectives_dict):
            k = 0
            if obj.component().sense == maximize:
                k = 1
            if ampl_repn is None:
                continue
            OUTPUT.write( "O%d %d\n"%(self_ampl_obj_id[obj_name], k) )
            if symbolic_solver_labels is True:
                rowf.write(name_labeler(obj)+"\n")

            if ampl_repn.is_linear():
                OUTPUT.write( "n{0!r}\n".format(ampl_repn._constant) )
            else:
                if ampl_repn._constant != 0.0:
                    # +
                    OUTPUT.write( "o0\nn{0!r}\n".format(ampl_repn._constant) )
                self._print_nonlinear_terms_NL(OUTPUT, ampl_repn._nonlinear_expr, symbol_map, labeler)

        if symbolic_solver_labels is True:
            rowf.close()
        del name_labeler

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing intializations:",
#        start_time = time.clock()

        #
        # "d" lines
        #
        # dual initialization
        if has_dual_initials is True:
            dual_suffix = model.dual
            s_lines = []
            for con_name, (con_component, ampl_repn) in iteritems(Constraints_dict):
                suffix_value = dual_suffix.getNumericValue(con_component)
                if (suffix_value is not None) and (suffix_value != 0):
                    s_lines.append(suffix_line.format(self_ampl_con_id[con_name],suffix_value))
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write( "d%d\n"%(len_s_lines) )
                [OUTPUT.write(line) for line in s_lines]
            

        #
        # "x" lines
        #
        # variable initialization
        var_bound_list = []
        x_init_list = []
        for ampl_var_id, var_name in enumerate(full_var_list):
            var = Vars_dict[var_name]
            # FIXME: This will need to change when var.value changes/goes away
            #        The test is here because we are inconsistent about initializing
            #        values of variables
            if var.value is not None:
                x_init_list.append("{0} {1!r}\n".format(ampl_var_id,var.value))
            if isinstance(var.domain, BooleanSet):
                var_bound_list.append("0 0 1\n")
                symbol_map.alias(var, ".v%s"%(ampl_var_id))
                continue
            L = var.lb
            U = var.ub
            if L is not None:
                Lv = value(L)
                if U is not None:
                    Uv = value(U)
                    if Lv == Uv:
                        var_bound_list.append("4 {0!r}\n".format(Lv))
                    else:
                        var_bound_list.append("0 {0!r} {1!r}\n".format(Lv, Uv))
                else:
                    var_bound_list.append("2 {0!r}\n".format(Lv))
            elif U is not None:
                var_bound_list.append("1 {0!r}\n".format(value(U)))
            else:
                var_bound_list.append("3\n")
            symbol_map.alias(var, ".v%d"%(ampl_var_id))


        OUTPUT.write( "x%d\n"%(len(x_init_list)) )
        [OUTPUT.write(line) for line in x_init_list]
        del x_init_list
        
#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing constraint bounds:",
#        start_time = time.clock()
        #
        # "r" lines
        #
        OUTPUT.write( "r\n" )
        # *NOTE: This iteration follows the assignment of the ampl_con_id
        [OUTPUT.write(constraint_bounds_dict[con_name]) for con_name in itertools.chain(nonlin_con_order_list, lin_con_order_list)]
        del constraint_bounds_dict

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing variable bounds:",
#        start_time = time.clock()
        #
        # "b" lines
        #
        OUTPUT.write( "b\n" )
        [OUTPUT.write(line) for line in var_bound_list]
        del var_bound_list
        
#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing k's:",
#        start_time = time.clock()        

        #
        # "k" lines
        #
        ktot = 0
        n1 = len(Vars_dict) - 1
        OUTPUT.write( "k%d\n"%(n1) )
        ktot = 0
        for i in xrange(n1):
            ktot += cu[i]
            OUTPUT.write( "%d\n"%(ktot) )
        del cu

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing Jacobian's:",
#        start_time= time.clock()

        #
        # "J" lines
        # 
        for nc, con_name in enumerate(itertools.chain(nonlin_con_order_list,lin_con_order_list)):
            con_data, ampl_repn = Constraints_dict[con_name]
            con_vars = set(ampl_repn._linear_terms_var)
            con_vars.update(ampl_repn._nonlinear_vars)
            ### GAH: This replaces the checks for constraint.trivial,
            ###      and ampl_repn.fixed. In these cases no active variables
            ###      will exist inside the ampl_repn. The purpose here is to have
            ###      constraints that become trivial due to fixed variables still
            ###      be written in the "C" section as a feasibility check for the fixed
            ###      variable values.
            if len(con_vars) > 0:
                OUTPUT.write( "J%d %d\n"%(nc, len(con_vars)) )
                for con_var in sorted(con_vars):
                    if con_var in ampl_repn._linear_terms_var:
                        OUTPUT.write( "{0} {1!r}\n".format(self_ampl_var_id[con_var], ampl_repn._linear_terms_coef[con_var]) )
                    else:
                        OUTPUT.write( "%d 0\n"%(self_ampl_var_id[con_var]) )

#        end_time = time.clock()
#        print (end_time - start_time)

#        print "Printing objective gradient:",
#        start_time = time.clock()
        #
        # "G" lines
        #
        no = 0
        grad_entries = {}
        for obj, ampl_repn in itervalues(Objectives_dict):
            obj_vars = set(ampl_repn._linear_terms_var)
            obj_vars.update(ampl_repn._nonlinear_vars)
            if len(obj_vars) == 0:
                pass
            else:
                OUTPUT.write( "G%d %d\n"%(no, len(obj_vars)) )
            no += 1

            for obj_var in obj_vars:
                if obj_var in ampl_repn._linear_terms_var:
                    grad_entries[self_ampl_var_id[obj_var]] = ampl_repn._linear_terms_coef[obj_var]
                else:
                    grad_entries[self_ampl_var_id[obj_var]] = 0

        for var_id in sorted(iterkeys(grad_entries)):
            OUTPUT.write( "{0} {1!r}\n".format(var_id, grad_entries[var_id]) )

#        print (end_time - start_time)
#        end_overall_time = time.clock()
#        print "Total time:",end_overall_time-start_overall_time

        return symbol_map


class ampl_representation:
    __slots__ = ('_constant', '_linear_terms_var', '_linear_terms_coef', '_nonlinear_expr', '_nonlinear_vars')
    def __init__(self):
        self._constant = 0
        self._linear_terms_var = {}
        self._linear_terms_coef = {}
        self._nonlinear_expr = None
        self._nonlinear_vars = {}

    def clone(self):
        clone = ampl_representation()
        clone._constant = self._constant
        clone._linear_terms_var = self._linear_terms_var
        clone._linear_terms_coef = self._linear_terms_coef
        clone._nonlinear_expr = self._nonlinear_expr
        clone._nonlinear_vars = self._nonlinear_vars
        return clone

    # for debugging
    def __str__(self):
        output = StringIO()
        output.write("\n")
        output.write("constant:       "+str(self._constant)+"\n")
        output.write("linear vars:    "+str(self._linear_terms_var)+"\n")
        output.write("linear coef:    "+str(self._linear_terms_coef)+"\n")
        output.write("nonlinear expr:\n")
        try:
            self._nonlinear_expr.pprint(ostream=output)
        except AttributeError:
            output.write(str(self._nonlinear_expr)+"\n")
        output.write("nonlinear vars: "+str(self._nonlinear_vars)+"\n")
        output.write("\n")
        ret_str = output.getvalue()
        output.close()
        return ret_str

    def __eq__(self, other):
        # Can only be equal to other ampl_representation instances
        if not isinstance(other, ampl_representation):
            return False
        
        # Immediately check constants
        if self._constant != other._constant:
            return False

        # Optimization: check linear term lengths before iterating
        if len(self._linear_terms_var) != len(other._linear_terms_var):
            return False

        # Establish a mapping between self's linear terms and other's
        linear_var_indices = {}
        for i in self._linear_terms_var:
            for j in other._linear_terms_var:
                if self._linear_terms_var[i] is other._linear_terms_var[j]:
                    linear_var_indices[i] = j

        # We have to have found all our own vars
        if len(linear_var_indices) != len(self._linear_terms_var):
            return False

        # Check that linear term coefficients are equal
        for i in self._linear_terms_var:
            if self._linear_terms_coef[i] != other._linear_terms_coef[linear_var_indices[i]]:
                return False

        if self._nonlinear_expr is not None:
            if other._nonlinear_expr is None:
                return False

            if type(self._nonlinear_expr) is list:
                if type(other._nonlinear_expr) is not list:
                    return False

                # Check that nonlinear expressions are the same length
                if len(self._nonlinear_expr) != len(other._nonlinear_expr):
                    return False

                # Check that nonlinear expression have the same terms
                for i in xrange(len(self._nonlinear_expr)):
                    if self._nonlinear_expr[i][0] != other._nonlinear_expr[i][0]:
                        return False
                    if not self._nonlinear_expr[i][1] is other._nonlinear_expr[i][1]:
                        return False
            else:
                if not self._nonlinear_expr is other._nonlinear_expr:
                    return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_fixed(self):
        if len(self._linear_terms_var) == 0 and len(self._nonlinear_vars) == 0:
            return True
        return False

    def is_linear(self):
        if self._nonlinear_expr is None:
            return True

    def is_nonlinear(self):
        if self._nonlinear_expr is None:
            return False
        return True

def generate_ampl_repn(exp, symbol_map, labeler):
    ampl_repn = ampl_representation()

    #
    # Expression
    #
    if exp.is_expression():#isinstance(exp,expr.Expression):
        exp_type = type(exp)

        #
        # Sum
        #
        if exp_type is expr._SumExpression:
            ampl_repn._constant = exp._const #float(exp._const)
            ampl_repn._nonlinear_expr = None
            for i in xrange(len(exp._args)):
                exp_coef = exp._coef[i]
                child_repn = generate_ampl_repn(exp._args[i], symbol_map, labeler)
                # adjust the constant
                ampl_repn._constant += exp_coef*child_repn._constant

                # adjust the linear terms
                for (var_name,var) in iteritems(child_repn._linear_terms_var):
                    if var_name in ampl_repn._linear_terms_var:
                        ampl_repn._linear_terms_coef[var_name] += exp_coef*child_repn._linear_terms_coef[var_name]
                    else:
                        ampl_repn._linear_terms_var[var_name] = var
                        ampl_repn._linear_terms_coef[var_name] = exp_coef*child_repn._linear_terms_coef[var_name]

                # adjust the nonlinear terms
                if not child_repn._nonlinear_expr is None:
                    if ampl_repn._nonlinear_expr is None:
                        ampl_repn._nonlinear_expr = [(exp_coef, child_repn._nonlinear_expr)]
                    else:
                        ampl_repn._nonlinear_expr.append((exp_coef, child_repn._nonlinear_expr))

                # adjust the nonlinear vars
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
            return ampl_repn

        #
        # Product
        #
        elif exp_type is expr._ProductExpression:
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.is_fixed():
                    denom *= value(e)
                else:
                    ampl_repn._nonlinear_expr = exp
                    break
                if denom == 0.0:
                    print("Divide-by-zero error - offending sub-expression:")
                    e.pprint()
                    raise ZeroDivisionError

            if not ampl_repn._nonlinear_expr is None:
                # we have a nonlinear expression ... build up all the vars
                for e in exp._denominator:
                    arg_repn = generate_ampl_repn(e, symbol_map, labeler)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)

                for e in exp._numerator:
                    arg_repn = generate_ampl_repn(e, symbol_map, labeler)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                return ampl_repn

            #
            # OK, the denominator is a constant.
            #
            # build up the ampl_repns for the numerator
            
            ###### GAH: Deactivating this preprocessing step, which 
            ######      also eliminates the need for expression cloning
            # preprocess out the denominator
            #exp._denominator = []
            #exp._coef *= 1.0/denom
            # Make sure to reset the denom to 1.0, this is important
            # when the outer expression is linear so that
            # we don't propigate this effect into the ampl
            # repn more than once.
            #denom = 1.0
            ######

            n_linear_args = 0
            n_nonlinear_args = 0
            arg_repns = list()
            for i in xrange(len(exp._numerator)):
                e = exp._numerator[i]
                e_repn = generate_ampl_repn(e, symbol_map, labeler)
                arg_repns.append(e_repn)
                # check if the expression is not nonlinear else it is nonlinear
                if e_repn._nonlinear_expr is not None:
                    n_nonlinear_args += 1
                # Check whether the expression is constant or else it is linear
                elif len(e_repn._linear_terms_var) > 0:
                    n_linear_args += 1
                # At this point we do not have a nonlinear expression and there are
                # no linear terms. If the expression constant is zero, then we
                # have a zero term in the product expression, so the entire product 
                # expression becomes trivial. 
                elif e_repn._constant == 0.0:
                    ampl_repn = e_repn
                    return ampl_repn

            is_nonlinear = False
            if n_linear_args > 1 or n_nonlinear_args > 0:
                is_nonlinear = True


            if is_nonlinear is True:
                # do like AMPL and simply return the expression
                # without extracting the potentially linear part
                ampl_repn = ampl_representation()
                ampl_repn._nonlinear_expr = exp
                for repn in arg_repns:
                    ampl_repn._nonlinear_vars.update(repn._linear_terms_var)
                    ampl_repn._nonlinear_vars.update(repn._nonlinear_vars)
                return ampl_repn

            else: # is linear or constant
                ampl_repn = current_repn = arg_repns[0]
                for i in xrange(1,len(arg_repns)):
                    e_repn = arg_repns[i]
                    ampl_repn = ampl_representation()

                    # const_c * const_e
                    ampl_repn._constant = current_repn._constant * e_repn._constant

                    # const_e * L_c
                    if e_repn._constant != 0.0:
                        for (var_name, var) in iteritems(current_repn._linear_terms_var):
                            ampl_repn._linear_terms_coef[var_name] = current_repn._linear_terms_coef[var_name] * e_repn._constant
                            ampl_repn._linear_terms_var[var_name] = var

                    # const_c * L_e
                    if current_repn._constant != 0.0:
                        for (e_var_name,e_var) in iteritems(e_repn._linear_terms_var):
                            if e_var_name in ampl_repn._linear_terms_var:
                                ampl_repn._linear_terms_coef[e_var_name] += current_repn._constant * e_repn._linear_terms_coef[e_var_name]
                            else:
                                ampl_repn._linear_terms_coef[e_var_name] = current_repn._constant * e_repn._linear_terms_coef[e_var_name]
                                ampl_repn._linear_terms_var[e_var_name] = e_var
                    current_repn = ampl_repn

            # now deal with the product expression's coefficient that needs
            # to be applied to all parts of the ampl_repn
            ampl_repn._constant *= exp._coef/denom
            for var_name in ampl_repn._linear_terms_coef:
                ampl_repn._linear_terms_coef[var_name] *= exp._coef/denom

            return ampl_repn

        #
        # Power Expressions
        #
        elif exp_type is expr._PowExpression:
            assert(len(exp._args) == 2)
            base_repn = generate_ampl_repn(exp._args[0], symbol_map, labeler)
            exponent_repn = generate_ampl_repn(exp._args[1], symbol_map, labeler)

            if base_repn.is_fixed() and exponent_repn.is_fixed():
                ampl_repn._constant = base_repn._constant**exponent_repn._constant
            elif exponent_repn.is_fixed() and exponent_repn._constant == 1.0:
                ampl_repn = base_repn
            elif exponent_repn.is_fixed() and exponent_repn._constant == 0.0:
                ampl_repn._constant = 1.0
            else:
                # instead, let's just return the expression we are given and only
                # use the ampl_repn for the vars
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = base_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(exponent_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(base_repn._linear_terms_var)
                ampl_repn._nonlinear_vars.update(exponent_repn._linear_terms_var)
            return ampl_repn
        #
        # External Functions
        #
        elif exp_type is external._ExternalFunctionExpression:
            # In theory, the argument are fixed, we can simply evaluate this expression
            if exp.is_fixed():
                ampl_repn._constant = value(exp)
                return ampl_repn

            # this is inefficient since it is using much more than what we need
            ampl_repn._nonlinear_expr = exp
            for arg in exp._args:
                if isinstance(arg, basestring):
                    continue
                child_repn = generate_ampl_repn(arg, symbol_map, labeler)
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(child_repn._linear_terms_var)
            return ampl_repn
        #
        # Intrinsic Functions
        #
        elif isinstance(exp,expr._IntrinsicFunctionExpression):
            # Checking isinstance above accounts for the fact that _AbsExpression
            # is a subclass of _IntrinsicFunctionExpression
            assert(len(exp._args) == 1)
            
            # the argument is fixed, we can simply evaluate this expression
            if exp._args[0].is_fixed():
                ampl_repn._constant = value(exp)
                return ampl_repn

            # this is inefficient since it is using much more than what we need
            child_repn = generate_ampl_repn(exp._args[0], symbol_map, labeler)

            ampl_repn._nonlinear_expr = exp
            ampl_repn._nonlinear_vars = child_repn._nonlinear_vars
            ampl_repn._nonlinear_vars.update(child_repn._linear_terms_var)
            return ampl_repn
        #
        # AMPL-style If-Then-Else expression
        #
        elif exp_type is Expr_if:
            if_repn = generate_ampl_repn(exp._if, symbol_map, labeler)
            if if_repn.is_fixed():
                if bool(if_repn._constant) is True:
                    ampl_repn = generate_ampl_repn(exp._then, symbol_map, labeler)
                else:
                    ampl_repn = generate_ampl_repn(exp._else, symbol_map, labeler)
            else:
                then_repn = generate_ampl_repn(exp._then, symbol_map, labeler)
                else_repn = generate_ampl_repn(exp._else, symbol_map, labeler)
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = if_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(then_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(else_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(if_repn._linear_terms_var)
                ampl_repn._nonlinear_vars.update(then_repn._linear_terms_var)
                ampl_repn._nonlinear_vars.update(else_repn._linear_terms_var)
            return ampl_repn
        elif (exp_type is expr._InequalityExpression) or (exp_type is expr._EqualityExpression):
            for arg in exp._args:
                arg_repn = generate_ampl_repn(arg, symbol_map, labeler)
                ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(arg_repn._linear_terms_var)
            ampl_repn._nonlinear_expr = exp
            return ampl_repn
        #
        # ERROR
        #
        else:
            raise ValueError("Unsupported expression type: "+str(exp))
    #
    # Constant
    #
    elif exp.is_fixed():
        ### GAH: Why were we even checking this
        #if not exp.value.__class__ in native_numeric_types:
        #    ampl_repn = generate_ampl_repn(exp.value, symbol_map, labeler) 
        #    return ampl_repn
        ampl_repn._constant = value(exp)
        return ampl_repn
    #
    # Variable
    #
    elif isinstance(exp, _VarData):
        if exp.fixed:
            ampl_repn._constant = exp.value
            return ampl_repn
        use_name = symbol_map.getSymbol( exp, labeler )
        ampl_repn._linear_terms_coef[use_name] = 1.0
        ampl_repn._linear_terms_var[use_name] = exp
        return ampl_repn

    #
    # ERROR
    #
    else:
        raise ValueError("Unexpected expression type: "+str(exp))



# Switch from Python to C generate_ampl_repn function when possible
try:
    py_generate_ampl_repn = generate_ampl_repn
    from cAmpl import generate_ampl_repn
except ImportError:
    del py_generate_ampl_repn

# Alternative: import C implementation under another name for testing
#from cAmpl import generate_ampl_repn as cgar
#__all__.append('cgar')
